import json
import inspect

from datetime import datetime
from functools import wraps


def profiled(f):
  """
  A decorator that records and prints the time taken for the decorated function to complete

  TODO: improve this (but this rudimentary format has proven useful for development)
  """

  calling_module = inspect.getmodule(inspect.currentframe().f_back)
  calling_module_name = calling_module and calling_module.__name__ or None
  @wraps(f)
  def run(*args, **kwargs):
    start = datetime.now()
    try:
      return f(*args, **kwargs)
    finally:
      end = datetime.now()
      print("[@profiled] " + json.dumps([calling_module_name, f.__name__, (end - start).total_seconds()]))
  return run



from threading import Lock


def singleton(cls):
  """
  A class decorator that ensures there can only ever be one instance of the decorated class. It prevents the decorated
  class from being created by overriding its '__new__' function, and it adds a new static zero-argument thread-safe
  function named 'instance' that should be used to initialize / retrieve the singleton instance. The wrapped class must
  have a zero-argument constructor, and it must not have an attribute named 'instance'.

  Example:
  @singleton
  class Test(object):
    pass

  Test() # Calling the constructor directly will raise an exception
  Test.instance() == Test.instance() # True
  type(Test.instance()) == Test # True
  Test.__name__ # Singleton<Test>
  """

  if hasattr(cls, 'instance'):
    raise Exception("The attribute 'instance' is reserved in classes decorated with '@singleton'")

  # We are going to replace these properties, but we still want references to the originals
  orig__new__ = cls.__new__
  orig__name__ = cls.__name__

  def override__new__(self):
    raise Exception("%s cannot be instantiated directly (use '%s.instance()' instead)" % (cls.__name__, orig__name__))
  cls.__new__ = override__new__
  cls.__name__ = "Singleton<%s>" % cls.__name__

  __instance = None
  __instance_lock = Lock()
  def instance():
    nonlocal __instance, __instance_lock
    if not __instance:
      __instance_lock.acquire()
      try:
        if not __instance:
          __instance = orig__new__(cls)
          __instance.__init__()
      finally:
        __instance_lock.release()
    return __instance
  cls.instance = instance

  return cls


def singleton_mock(singleton_cls):
  """
  Example:

  @singleton
  class Foo(object):
    def bar(self):
      return "a"
    def baz(self):
      return "b"

  Foo.instance().bar() -> "a"

  @singleton_mock(Foo)
  class MockFoo(object):
    def bar(self):
      return "c"

  Foo.instance().bar() -> "c"
  Foo.instance().baz() -> Exception

  # You can do partial mocks by extending from the original object
  @singleton_mock(Foo)
  class MockFoo(Foo):
    def baz(self):
      return "d"

  Foo.instance().bar() -> "a"
  Foo.instance().baz() -> "d"
  """

  if not hasattr(singleton_cls, 'instance'):
    raise Exception("The class '%s' does not appear to be decorated with '@singleton'" % singleton_cls)

  def replace_with_class(mock_cls):
    singleton_mock_cls = singleton(mock_cls)
    singleton_cls.instance = singleton_mock_cls.instance
    singleton_cls.__name__ = singleton_mock_cls.__name__
    return singleton_mock_cls
  return replace_with_class



import inspect
import os

from functools import wraps

from .exceptions import retry


def override_eq(f):
  """
  A decorator that converts the wrapped function into a class that will be used to determine if two arguments are equal
  when it is used as the left-side operand of an '==' expression. This can be useful for functions / statements that
  compare against an operand rather than allowing a custom comparator function.

  Paramters:
  - f: a single-argument function that will be passed the right side operand when this decorator is used in an equality
    comparison operation

  See the function '__eq_any__' for an example.

  Examples:

  @override_eq
  def is_positive(n):
    return n > 0
  is_positive == 1 # True
  is_positive == -1 # False

  override_eq(lambda i: i > 0) == 1 # True
  override_eq(lambda i: i > 0) == -1 # False
  """

  class OverrideEquals(object):

    def __eq__(self, other):
      return f(other)

  return OverrideEquals()


@override_eq
def __eq_any__(right_side):
  """
  A class that always returns True in an equality test (if it is the left argument)

  Examples:
  __eq_any__ == 1 # True
  __eq_any__ == 2 # True
  (1, __eq_any__) == (1, 2) # True
  """

  return True


class EventualAssertionError(AssertionError):
  pass


def assert_eventual(expected_output, delay_s=0.5, max=10, exp_backoff=1):
  """
  A decorator that retries the wrapped function until its output matches the 'expected_output'. If the retry conditions
  are exhausted, this function throws an EventualAssertionError (subclass of AssertionError), otherwise it returns the
  function output as normal.

  Note that you can also use this function inline to replace an assert statement:
  assert eventually_consistent_test_func() == 5 # May fail if the function takes a bit to return the expected
  assert_eventual(5)(eventually_consistent_test_func)() # Automatically retries

  Note that 'expected_output' will always be used as the left side operator when comparing equality against the output,
  so you could use a custom comparator class that overrides the '__eq__' function to do partial matching (see
  'override_eq' function).

  Parameters:
  - expected_output: the value against which to compare the actual function output

  Examples:

  i = 0
  @assert_eventual(5, delay_s=0)
  def test():
    global i
    i += 1
    return i
  test() # Returns 5 after running five times (with four retries)

  @assert_eventual(override_eq(lambda right_side: right_side['a'] == 1))
  def test():
    return { 'a': 1, 'b': 2 }
  test() # Returns { 'a': 1, 'b': 2 }

  assert_eventual(5)(lambda: 5)() # Returns 5
  """

  def decorator(f):
    @retry(EventualAssertionError, delay_s=delay_s, max=max, exp_backoff=exp_backoff)
    @wraps(f)
    def run(*args, **kwargs):
      output = f(*args, **kwargs)
      if expected_output != output:
        raise EventualAssertionError(output)
      return output
    return run
  return decorator


def exec_module_at_path(path):
  """
  Run the code in a module at the given path (either a file or directory containing an __init__.py), but don't import
  it into our current namespace. Relative paths will be resolved using the path of the calling function's module.

  Warning: unlike python import statements, this function does not prohibit circular execution.

  Example:

  /module_a.py
    print("module_a")

  /module_b/__init__.py
    print("module_b")

  /test_1.py
    load_path("module_a")
    load_path("module_b")

  /dir/test_2.py
    load_path("../module_a")
    load_path("../module_b")

  /test_3.py
    load_path("module_a")

  # Note that it doesn't matter from what directory I run these examples, since the path is relative to the module in
  # which the function is called
  $ cd / && python test_1.py
  > module_a
  > module_b
  $ cd /dir && python /test_1.py
  > module_a
  > module_b
  $ cd /dir && python test_2.py
  > module_a
  > module_b
  $ python /test_3.py
  > Exception
  """

  frame = inspect.stack()[1]
  module = inspect.getmodule(frame[0])
  dir = os.path.dirname(module.__file__)
  path = os.path.join(dir, path)

  if not os.path.isfile(path):
    path = os.path.join(path, "__init__.py")
  if not os.path.isfile(path):
    raise Exception("Module not found at path '%s'" % path)
  with open(path) as file:
    exec(file.read())



from op_service_client import OPJSONEncoder
from starlette.datastructures import UploadFile


# Use this encoder for printing output as encoded by the OPJSONEncoder - it makes special accommodations for
# UploadFiles and handles any error by defaulting to str
class OPJSONEncoderFallback(OPJSONEncoder):

  def default(self, obj):
    if isinstance(obj, UploadFile):
      return 'File<%s, %s>' % (obj.filename, obj.content_type)
    try:
      return super().default(obj)
    except TypeError:
      return str(obj)



import asyncio
import functools
import importlib.util
import inspect
import json
import logging
import op_alarm.factory
import op_storage.factory
import os
import re
import signal
import socket
import sys
import traceback

from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager, suppress
from datetime import datetime, timezone
from op_service_client import OPJSONEncoder
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import Response
from threading import Condition, Thread

from .env import OP_DEV_CONFIG, OP_DEV_METADATA
from .exceptions import AuthenticationException, ClientException, InvalidAuthorizationHeader, InvalidInput, ServiceException, ignore_exceptions
from .instance_id import SERVICE_INSTANCE_ID
from .jwt_manager import JWTManager
from .logging_config import get_relative_sys_filename
from .op_json_encoder import OPJSONEncoderFallback
from .task_manager import OPTaskManager
from .type_validation import describe, validate


_REQUEST_LOGGER = logging.getLogger('op-request')

# This structure describes the order / priority for printing out request state information - the lists are read from top
# to bottom, and keys that come first in a list will be printed instead of any following keys.
_STATE_GROUPS = [
  [ 'profile_validated', 'profile' ],
  [ 'input_data_validated', 'input_data' ]
]

# key -> (header, transform)
_STATE_KEY_OUTPUTS = {
  'profile': 'p', # Input profile
  'profile_validated': 'pv', # Input profile (validated)
  'input_data': 'i', # Input data
  'input_data_validated': 'iv' # Input data (validated)
}

def _on_route_start(request_state):
  _REQUEST_LOGGER.info(json.dumps({ '_': 0, 'r': request_state['route'], 'ip': request_state['ip'], 'ua': request_state['ua'] }, separators=(',', ':')))

def _on_route_end(request_state):
  msg = { '_': 1, 'r': request_state['route'], 'ip': request_state['ip'], 'ua': request_state['ua'], 'c': request_state['response_code'] }
  for state_group in _STATE_GROUPS:
    for key in state_group:
      if key in request_state:
        msg[ _STATE_KEY_OUTPUTS[key] ] = request_state[key]
        break
  if 'o' in msg:
    # This is a special case: this output is already serialized
    del msg['o']
  packed = json.dumps(msg, cls=OPJSONEncoderFallback, separators=(',', ':'))
  if 'response_data' in request_state:
    # The 'response_data' is a special case because it is already serialized (and potentially large), so we can avoid
    # reserializing it if we just inject it directly
    packed = packed[:-1] + ',"o":' + request_state['response_data'] + '}'
  _REQUEST_LOGGER.info(packed)


_IP_ADDR_V4_REGEX = re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}')

def _get_ip_address(request):
  # In non-test environments, all requests should have the X-Forwarded-For header set correctly by our proxy
  x_forwarded_for = request.headers.get('X-Forwarded-For')
  if not x_forwarded_for:
    return request.client.host
  # https://docs.aws.amazon.com/elasticloadbalancing/latest/application/x-forwarded-headers.html#x-forwarded-for
  with suppress(Exception):
    return _IP_ADDR_V4_REGEX.match(x_forwarded_for)[0]
  # This must be an IP V6 address - AWS may wrap the V6 address in brackets if it includes a port
  if x_forwarded_for.startswith('['):
    ip_v6 = x_forwarded_for.split(']', 1)[0][1:]
  else:
    ip_v6 = x_forwarded_for.split(',', 1)[0]
  with suppress(Exception):
    socket.inet_pton(socket.AF_INET6, ip_v6)
    return ip_v6
  # We couldn't determine the IP address from the provided string. This shouldn't happen as all requests must have an
  # IP address, but I'm not confident enough in my logic / configuration yet to throw an exception in this case.
  return ''


# The allowed argument names for functions decorated with @route
_ALLOWED_ARGS = [ 'request', 'ip_address', 'user_agent', 'profile', 'data' ]

class OPService:

  def __init__(self):
    # Note: I generally don't like the concept of "middleware" as it's less clear where / how this code runs.
    # Generally, I prefer to explicitly wrap "middleware" around the "run" function inside the "route" decorator,
    # except where Starlette's built-ins are super convenient...
    middleware = [ Middleware(CORSMiddleware, allow_origins=[ '*' ], allow_headers=[ '*' ], allow_methods=[ '*' ]) ]

    self.starlette = Starlette(middleware=middleware, on_startup=[ self._on_startup ])

    self._init_thread_pool = ThreadPoolExecutor(max_workers=64)
    self._request_thread_pool = ThreadPoolExecutor(max_workers=64)
    self._routes = []
    self._init_fns = []
    self._init_condition = Condition()
    self._init_complete = False

    self.alarm = op_alarm.factory.build(OP_DEV_CONFIG['op_alarm']['type'], OP_DEV_CONFIG['op_alarm']['config'])
    self.db = op_storage.factory.build(OP_DEV_CONFIG['op_storage']['type'], OP_DEV_CONFIG['op_storage']['config'])
    self.jwt = JWTManager(OP_DEV_CONFIG['jwt'])
    self.task = OPTaskManager(self)

    # Every production service has a '/' health-check endpoint
    @self.route('/', methods=[ 'GET' ], private=True)
    async def health(request):
      with self._init_condition:
        self._init_condition.wait_for(lambda: self._init_complete)
      return Response('')

    # TODO: since these three APIs should only ever be called from the context of the op-dev dashboard, you should add
    # some corresponding authentication

    # Returns the commit / config IDs
    @self.route(path='/__private/metadata', private=True)
    async def metadata():
      return OP_DEV_METADATA

    # Returns documentation about the routes
    @self.route(path='/__private/explore', private=True)
    async def explore():
      grouped_routes = defaultdict(list)
      for route in self._routes:
        if not route['private']:
          grouped_routes[route['grouping']].append(route)
      output = []
      for key in sorted(grouped_routes):
        output.append(
          {
            'group': key,
            'routes': [
              {
                'name': x['name'],
                'profile': describe(x['profile_fmt']),
                'data': describe(x['data_fmt']),
                'methods': x['methods'],
                'documentation': x['fn'].__doc__
              } for x in grouped_routes[key]
            ]
          }
        )
      return output

    if os.environ.get('OP_DEV_SERVICE_TEST', False):
      # This service was brought up in dev mode (using 'op-dev run')
      # Add a private API that allows us to generate custom JWTs
      @self.route(path='/__private/generate-jwt', data=dict, private=True)
      async def generate_jwt(data):
        return self.jwt.encode(data)


  def _get_input_profile(self, request, profile_format, request_state):
    headers = request.headers
    auth_header_value = headers.get('Authorization')
    if auth_header_value:
      if not auth_header_value.startswith('Bearer '):
        raise InvalidAuthorizationHeader("Authorization header is malformatted - does not start with 'Bearer '")
      token = auth_header_value[7:]
      profile = self.jwt.decode(token)
    else:
      profile = None
    request_state['profile'] = profile
    try:
      request_state['profile_validated'] = validated = validate(profile, profile_format)
      return validated
    except Exception as e:
      raise AuthenticationException() from e


  async def _get_input_data(self, request, data_format, request_state):
    input_data = None
    if data_format is not None:
      content_type = request.headers.get('content-type', '')
      if content_type == 'application/json':
        body = await request.body()
        if body:
          try:
            input_data = json.loads(body)
          except Exception as e:
            raise ClientException('Error parsing JSON') from e
      elif content_type.startswith('multipart/form-data'):
        try:
          form = await request.form()
          def inject_files(d):
            file_id = len(d) == 1 and d.get('__file__')
            return form[file_id] if file_id else d
          input_data = json.loads(form['_'], object_hook=inject_files)
        except Exception as e:
          raise ClientException('Error parsing form data') from e
      else:
        raise ClientException('Unrecognized content-type')
    request_state['input_data'] = input_data
    try:
      request_state['input_data_validated'] = validated = validate(input_data, data_format)
      return validated
    except Exception as e:
      raise InvalidInput(e)


  # Functions with this decorator can still be called directly, though they won't get any input / profile validation
  def route(self, path=None, data=None, profile=None, methods=[ 'POST' ], private=False):
    data_fmt = data
    profile_fmt = profile

    def decorator(fn):
      nonlocal path, data_fmt, profile_fmt, methods, private

      route_name = path or ('/' + fn.__name__)
      for route in self._routes:
        if route['name'] == route_name:
          raise Exception(f'Duplicate route: {route_name}')

      calling_module = inspect.getmodule(inspect.currentframe().f_back)
      calling_module_name = calling_module and calling_module.__name__ or None
      self._routes.append({
        'name': route_name,
        'fn': fn,
        'profile_fmt': profile_fmt,
        'data_fmt': data_fmt,
        'methods': methods,
        'private': private,
        'grouping': calling_module_name
      })

      args = set(inspect.signature(fn).parameters)
      for arg in args:
        if arg not in _ALLOWED_ARGS:
          raise Exception('Unexpected parameter name for route function (allowed values: %s)' % _ALLOWED_ARGS)

      kwargs = { x: None for x in args }

      @self.starlette.route(route_name, methods=methods)
      async def run(request):
        ip_address = _get_ip_address(request)
        user_agent = request.headers.get('User-Agent')
        request_state = { 'route': route_name, 'ip': ip_address, 'ua': user_agent }

        # TODO: if the called URL is different than the route name, include it in the request_state (i.e. query string
        # parameters on a GET request)
        # if not request.url.endswith(route_name):
        #   request_state['route_actual'] = request.url

        _on_route_start(request_state)

        try:
          if profile_fmt:
            profile = self._get_input_profile(request, profile_fmt, request_state)
          if data_fmt:
            data = await self._get_input_data(request, data_fmt, request_state)
          for arg in args:
            # We've already validated that 'args' does not contain anything not in '_ALLOWED_ARGS', so this is safe
            kwargs[arg] = eval(arg)

          if asyncio.iscoroutinefunction(fn):
            output = await fn(**kwargs)
          else:
            # Make sure we don't block the main loop from accepting more requests. Note that bascially eliminates any
            # advantage we get using Starlette since everything is ultimately run in a thread anyway...
            output = await asyncio.get_event_loop().run_in_executor(self._request_thread_pool, functools.partial(fn, **kwargs))

          if issubclass(type(output), Response):
            request_state['response_type'] = type(output)
            request_state['response_code'] = output.status_code
            return output

          request_state['response_data'] = content = json.dumps(output, cls=OPJSONEncoder, separators=(',', ':'))
          request_state['response_code'] = 200
          return Response(content=content, media_type='application/json')

        except Exception as e:
          # Make it harder to run a DOS attack by just repeatedly calling APIs with bad input
          # await asyncio.sleep(1)

          request_state['exc_info'] = sys.exc_info()
          request_state['response_code'] = status_code = e.status_code if isinstance(e, ServiceException) else 500
          if status_code == 500:
            logging.exception('Uncaught exception on route: ' + route_name)
            with ignore_exceptions(log_msg='Alarm failed'):
              self._report(request_state)
            return Response(content='', status_code=500)
          request_state['response_data'] = content = json.dumps({ 'err': e.__class__.__name__, 'msg': str(e) }, separators=(',', ':'))
          return Response(content=content, status_code=status_code, media_type='application/json')

        finally:
          _on_route_end(request_state)

      return fn
    return decorator



  def init(self, block=True):
    def decorator(fn):
      if self._init_complete:
        raise Exception('Service is already initialized')
      self._init_fns.append((fn, block))
      return fn
    return decorator


  def _on_startup(self):
    logging.info('Starting worker...\nDomain: %(domain)s\nPackage ID: %(pkg_id)s\nCommit ID: %(commit_id)s\nConfig ID: %(config_id)s' % OP_DEV_METADATA)
    tasks = []
    for fn, block in self._init_fns:
      if asyncio.iscoroutinefunction(fn):
        task = self._init_thread_pool.submit(functools.partial(asyncio.run, fn()))
      else:
        task = self._init_thread_pool.submit(fn)
      tasks.append((task, block))
    for task, block in tasks:
      if block:
        task.result()
    self._init_complete = True
    with self._init_condition:
      self._init_condition.notify_all()
    logging.info('Finished starting worker')

    if 'OP_DEV_SERVICE_MAIN' in os.environ:
      sys.argv = [ '__main__', *json.loads(os.environ['OP_DEV_SERVICE_MAIN']) ]
      spec = importlib.util.spec_from_file_location('__main__', '/var/www/app/__main__.py')
      module = importlib.util.module_from_spec(spec)
      sys.modules['__main__'] = module
      try:
        spec.loader.exec_module(module)
      except Exception as e:
        print(traceback.format_exc(), end='', flush=True)
      finally:
        signal.raise_signal(signal.SIGTERM)

    for task, block in tasks:
      if not block:
        task.result()
    self._init_thread_pool.shutdown()


  def _report(self, request_state):
    # Any 500-level error surfaced here should have 'exc_info' attached
    exc_type, exc_value, exc_traceback = request_state['exc_info']
    stack_summary = traceback.extract_tb(exc_traceback)
    cause = stack_summary[-1]
    rel_file = get_relative_sys_filename(cause.filename)

    msg = '%s -> %s - %s %s#%s' % (request_state['route'], request_state['response_code'], exc_type.__name__, rel_file, cause.lineno)
    now = datetime.now(timezone.utc)
    issue = {
      'exception': {
        'msg': repr(exc_value),
        'traceback': stack_summary.format()
      },
      'request': {
        'route': request_state['route'],
        'profile': request_state.get('profile_validated'),
        'input_data': request_state.get('input_data'),
        'response_code': request_state['response_code'],
        'response_data': request_state.get('response_data')
      },
      'metadata': OP_DEV_METADATA,
      'count': {
        'n': 1,
        't_start': now.isoformat(),
        't_end': now.isoformat()
      }
    }

    for key in [ 'profile', 'input_data', 'response_data' ]:
      if issue['request'][key]:
        issue['request'][key] = json.dumps(issue['request'][key], cls=OPJSONEncoderFallback)

    self.alarm.report_issue(msg, **issue)


  @contextmanager
  def report_exceptions(self, msg):
    kwargs = { 'metadata': OP_DEV_METADATA }
    try:
      yield
    except Exception:
      logging.exception(msg)
      try:
        _, exc_value, exc_traceback = sys.exc_info()
        stack_summary = traceback.extract_tb(exc_traceback)
        kwargs['exception'] = { 'msg': repr(exc_value), 'traceback': stack_summary.format() }
        self.alarm.report_issue(msg, **kwargs)
      except:
        logging.exception('Failed to report exception')


service = OPService()



import json
import logging
import logging.config
import os
import sys

from .cache import cache
from .instance_id import SERVICE_INSTANCE_ID


@cache()
def get_relative_sys_filename(pathname):
  sys_paths = sorted(sys.path, key=len, reverse=True)
  if not sys_paths[-1]:
    sys_paths[-1] = os.getcwd()
  for sys_path in sys_paths:
    if sys_path and pathname.startswith(sys_path):
      return os.path.relpath(pathname, sys_path)
  return pathname


class OPFormatter(logging.Formatter):

  def format(self, record):
    relative_filename = get_relative_sys_filename(record.pathname)
    json_contents = {
      'l': record.levelname,
      'n': record.name,
      's': relative_filename,
      '#': record.lineno,
      'i': SERVICE_INSTANCE_ID,
      'm': super().format(record)
    }
    return "[OP]" + json.dumps(json_contents, separators=(',', ':'))


OP_LOG_CONFIG = {
  'version': 1,
  'disable_existing_loggers': False,
  'formatters': {
    'op_formatter': {
      '()': OPFormatter
    }
  },
  'handlers': {
    'default': {
      'formatter': 'op_formatter',
      'class': 'logging.StreamHandler'
    },
  },
  'root': {
    'handlers': [ 'default' ],
    'level': 'NOTSET'
  }
}


def configure_logging():
  logging.config.dictConfig(OP_LOG_CONFIG)
  logging.getLogger('fasteners.process_lock').setLevel(logging.WARNING)
  logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
  logging.getLogger('multipart.multipart').setLevel(logging.WARNING)
  logging.getLogger('hypercorn.access').setLevel(logging.WARNING)
  logging.getLogger('hypercorn.error').setLevel(logging.WARNING)



import logging
import time

from .exceptions import ThrottledException


class ThrottleMonitor:

  def __init__(self, config):
    self._n = config['n']
    self._period = config['period']
    self._reporter = config['reporter']
    self._state = {}
    self._blocked = set()

  async def block_ip(self, ip):
    if ip and (ip not in self._blocked):
      logging.critical('Blocking IP address due to call rate: %s' % ip)
      self._blocked.add(ip)
      await self._reporter.block_ip(ip)

  def check(self, ip):
    if ip in self._blocked:
      raise ThrottledException()

    # Keep the state clean
    now = time.time()
    old = []
    for ip, state in self._state.items():
      if now - state['t'] > self._period:
        old.append(ip)
      else:
        # Dictionaries preserve insertion order, so we know all the remaining records are within the period
        break
    for ip in old:
      del self._state[ip]

    if ip not in self._state:
      self._state[ip] = { 't': now, 'n': 0 }
    self._state[ip]['n'] += 1

    if self._state[ip]['n'] > self._n:
      raise ThrottledException()



import logging
from functools import wraps
from threading import Lock, Semaphore, Thread


DEFAULT_THREAD_POOL_SIZE = 64


# TODO: this might be a more useful interface:
# a, b = gather(Thread(target=fn_a, args=(), kwargs={}), Thread(target=fn_b, args=(), kwargs={}), max_n_threads=None)


def op_async(logger=None, thread_pool_size=DEFAULT_THREAD_POOL_SIZE):
  """
  A decorator that users the python 'Thread' object to run tasks asynchronously - note that because of the GIL, this
  decorator is only useful for non-compute tasks, like network calls. A call to the decorated function will return a
  special subclass of the thread object containing an 'wait' function. Use this function to get the results of the
  asynchronous call (or throw the exception if there is one).

  Example 1 (will complete in ~2 seconds because of the limitation on the thread_pool_size):
  @op_async(thread_pool_size=2)
  def test(i):
    import time
    time.sleep(1)
    return i
  threads = [test(i) for i in range(3)]
  results = [thread.wait() for thread in threads] # [0, 1, 2]

  Example 2 (throws an exception only when 'wait' is called):
  @op_async()
  def test():
    raise Exception()
  thread = test()
  thread.wait() # Exception is not raised until this line
  """

  if (type(thread_pool_size) is not int) or (thread_pool_size < 1):
    raise Exception("thread_pool_size must be >= 1")

  semaphore = Semaphore(thread_pool_size)
  def decorator(f):
    @wraps(f)
    def run(*args, **kwargs):
      
      class AwaitThread(Thread):

        def run(self):
          semaphore.acquire()
          try:
            self.result = f(*args, **kwargs)
          except Exception as e:
            if logger:
              logger.exception("Encountered exception in 'op_async' task")
            self.exception = e
          finally:
            semaphore.release()

        def wait(self):
          self.join()
          try:
            return self.result
          except:
            raise self.exception from None

      thread = AwaitThread()
      thread.start()
      return thread
    return run
  return decorator


def async_group(*wait_on):
  """
  Uses the 'op_async' pattern to execute the functions passed via the 'wait_on' parameter simultaneously. Returns a
  generator that delivers the result of the next sequential function as soon as it is available, or throws an exception.

  Parameters:
    'wait_on' is a list of zero-argument callables

  Example 1:
  import time
  def zero_argument_callable():
    time.sleep(1)
    return time.time()
  get = wait_group(zero_argument_callable, zero_argument_callable)
  for result in gen:
    print(result)

  Example 2 (if you don't care about the results):
  get = wait_group(zero_argument_callable, zero_argument_callable)
  tuple(gen)
  """

  threads = []
  for f in wait_on:
    thread = op_async()(f)()
    threads.append(thread)
  for thread in threads:
    yield thread.wait()


def synchronized(f):
  lock = Lock()
  def run(*args, **kwargs):
    lock.acquire()
    try:
      return f(*args, **kwargs)
    finally:
      lock.release()
  return run



import logging
import sys
import time

from contextlib import contextmanager, suppress
from functools import wraps


class ServiceException(Exception):
  status_code = 500


class NotFound(ServiceException):
  status_code = 404


class ClientException(ServiceException):
  status_code = 400


class InvalidInput(ClientException):
  pass


class AuthenticationException(ClientException):
  pass


class AuthorizationException(ServiceException):
  status_code = 401


class ConflictException(ServiceException):
  status_code = 409


class InvalidAuthorizationHeader(AuthorizationException):
  pass


class InvalidTokenSignature(AuthorizationException):
  pass


class ExpiredToken(AuthorizationException):
  pass


class ThrottledException(ServiceException):
  status_code = 429


def _is_match(exc, exc_type_handler):
  with suppress(TypeError):
    return issubclass(type(exc), exc_type_handler)
  return exc_type_handler(exc)


@contextmanager
def ignore_exceptions(*exc_type_handlers, log_msg=None):
  """
  Use this decorator / context manager to ignore specific exceptions. Non-ignored exceptions will be raised as
  normal.

  Parameters:
  - exc_type_handlers: (variable arity) specific exception types to ignore - either an exception class
    (subclass of BaseException, includes sub-classes), or a single-argument callable that will be passed the
    exception instance and should return True if it is a match

  Note the difference from 'contextlib.suppress' is the ability to use callables as exc_types.

  Examples:

  with ignore_exceptions(TypeError, lambda e: e.code == '404'):
    ...

  @ignore_exceptions(TypeError, lambda e: e.code == '404')
  def might_raise_exceptions():
    ...
  """

  try:
    yield
  except Exception as e:
    should_raise = bool(exc_type_handlers)
    for exc_type_handler in exc_type_handlers:
      if _is_match(e, exc_type_handler):
        should_raise = False
    if should_raise:
      raise
    if log_msg is not None:
      logging.exception(log_msg)


def retry(*exc_type_handlers, max=3, delay_s=0.1, exp_backoff=1.1):
  """
  Use this decorator to retry exceptions. By default, it retries all exceptions; use the 'exc_type_handlers'
  parameter to be more restrictive. If the retries are exhausted, the exception from the first execution will
  be raised.

  Parameters:
  - exc_type_handlers: (variable arity) a list of the specific exception types to retry - either an exception
    class (subclass of BaseException, includes sub-classes), or a single-argument callable that will be passed
    the exception instance and should return True if it is a match. Note that you can use these callables to
    do work before retrying the call, making it easy to implement a try, except, resolve, try-again pattern
    (the exception type handler can do work with side-effects to resolve the exception).
  - max: the max number of times to retry the call
  - delay_s: the initial delay between retries (delays between successive retry calls are calculated by
    multiplying the current 'delay_s' by the 'exp_backoff')
  - exp_backoff: the multiplier to increase the delay after the first retry

  Examples:

  @retry()
  def test():
    print("Will retry any exception 'max' times")
    raise Exception()

  @retry(TypeError)
  def test():
    print("Will not retry")
    raise Exception()

  should_raise = True
  def should_retry(e):
    global should_raise
    print("Resolving the exception as a side-effect")
    should_raise = False
    return True
  @retry(should_retry, max=1, delay_s=0)
  def test():
    if should_raise:
      print("Will retry after resolving the exception in the handler")
      raise Exception()
  """

  def can_retry(e):
    if not exc_type_handlers:
      return True
    for exc_type_handler in exc_type_handlers:
      if _is_match(e, exc_type_handler):
        return True
    return False

  def decorator(f):
    @wraps(f)
    def run(*args, **kwargs):
      i = 1
      cur_delay_s = delay_s
      first_exc = None

      while True:
        try:
          return f(*args, **kwargs)
        except Exception as e:
          if not can_retry(e):
            raise
          if not first_exc:
            # Preserve the original stack trace so it is not muddied by our logic here
            first_exc = sys.exc_info()
          if (i > max):
            break

        # TODO: if applicable, this log line should indicate what handler matched the exception (at least by
        # index) - successive retries may result in different exceptions, but only the first exception will be
        # raised
        msg = "Call to function '%s' raised exception - will retry after %.2fs delay (will be attempt #%i)..." % (f.__name__, cur_delay_s, i + 1)
        logging.info(msg, exc_info=True)

        time.sleep(cur_delay_s)
        i += 1
        cur_delay_s *= exp_backoff

      if first_exc:
        # We exhausted our retries, now time to raise the first exception
        exc_class, exc, tb = first_exc
        raise exc_class.with_traceback(exc, tb)

    return run
  return decorator



import setuptools
import sys

assert sys.version_info >= (3, 8)

setuptools.setup(
  name='op-service',
  version='0.1.6',
  author='Alex Exarhos',
  author_email='AlexExarhos@op-dev.io',
  packages=[
    'op_service'
  ],
  install_requires=[
    'iso8601 >= 0.1.11',
    'op-alarm >= 1.0.0',
    'op-service-client >= 0.1.1',
    'op-storage >= 1.0.0',
    'PyJWT >= 2.8.0',
    'python-multipart >= 0.0.6',
    'starlette >= 0.31.1'
  ]
)



'''
Defines a TaskManager build on top of an op_storage DB instance

TODO:
- Documentation
- I'm worried about going overboard with threads - how about we just report the heartbeats from a single thread?



Simplify submission to the thread pool executors
https://www.aeracode.org/2018/02/19/python-async-simplified/
def get_chat_id(name):
    return Chat.objects.get(name=name).id

async def main():
    executor = concurrent.futures.ThreadPoolExecutor(max_workers=3)
    loop = asyncio.get_event_loop()
    result = await loop.run_in_executor(executor, get_chat_id, "django")
'''

import asyncio
import atexit
import inspect
import logging
import pickle
import sys
import threading
import time
import traceback

from concurrent.futures import ThreadPoolExecutor
from concurrent.futures.thread import _python_exit as _concurrent_futures_thread_python_exit
from contextlib import contextmanager, suppress
from datetime import datetime, timedelta, timezone
from op_storage import Field
from op_storage.exceptions import NotFoundException
from threading import Event, Lock, RLock, Thread

from .env import OP_DEV_METADATA

# Unregister the exit handler associated with the ThreadPoolExecutor - the entire purpose of our library is to handle
# interrupted tasks, so we want those threads to behave like true daemons.
# See this for more info:
# https://stackoverflow.com/questions/49992329/the-workers-in-threadpoolexecutor-is-not-really-daemon
# WARNING: this changes the default behavior of 'concurrent.futures.thread' (for all consumers)
atexit.unregister(_concurrent_futures_thread_python_exit)


logger = logging.getLogger(__name__)


MIN_HEARTBEAT_S = 25
MAX_HEARTBEAT_S = 600
HEARTBEAT_EXP_GROWTH_FACTOR = 10

# We want to set the deadline for a task to exactly when the next heartbeat should occur (which in turn updates that
# deadline), but the timing won't be exact. Adding this buffer helps ensure that we will never encounter
# properly-running tasks that have missed their deadline.
HEARTBEAT_BUFFER_S = 10

# The manager polls the DB looking for abandoned tasks this often.
ABANDONED_POLL_INTERVAL_S = 30


def _now():
  return datetime.now(timezone.utc)


def _exp(i, growth_factor, min_value, max_value):
  # For computing exponential growth from the 'min_value' up to the 'max_value'
  return min(((2 ** i) -1) * growth_factor + min_value, max_value)


def _build_abandoned_task_msg(task):
  msg = (
    'Found abandoned task with unrecognized type - you need to spin up a worker that knows how to handle this task '
    'type'
  )
  task_metadata = task.get('source')
  if task_metadata and all(task_metadata.get(k) for k in ('pkg_id', 'commit_id', 'config_id')):
    msg += (
      '. You can do so by running the following command: op-dev run --pkg-id %(pkg_id)s --commit-id %(commit_id)s '
      '--config-id %(config_id)s'
    ) % task_metadata
  else:
    msg += ': ' + task['task_type_id']
  return msg


def _generate_task_type_id(fn):
  # Derives a task type ID from the given function automatically
  return f'{fn.__name__}{inspect.signature(fn)}'


class OPTaskException(Exception):
  pass


class TaskResult:

  def __init__(self, task):
    self._task = task

  def result(self):
    if 'output' in self._task:
      return self._task['output']
    raise OPTaskException(self._task['exc'])


class LocalTaskResult:

  def __init__(self, task, task_completed_event):
    self._task = task
    self._task_completed_event = task_completed_event

  def result(self):
    self._task_completed_event.wait()
    if 'output' in self._task:
      return self._task['output']
    raise OPTaskException(self._task['exc'])


class DistributedTaskResult:

  def __init__(self, task_run_id, manager, heartbeat_s):
    self._task_run_id = task_run_id
    self._m = manager
    self._heartbeat_s = heartbeat_s # Potentially None

  def result(self):
    i = 0
    while True:
      # We base our polling frequency on the heartbeat
      wait_s = self._heartbeat_s or _exp(i, HEARTBEAT_EXP_GROWTH_FACTOR, MIN_HEARTBEAT_S, MAX_HEARTBEAT_S)
      time.sleep(wait_s)
      try:
        task = self._m._service.db.get(self._m._data_type, self._task_run_id)
      except Exception as e:
        # Just adding amore descriptive error message
        raise Exception('Failed polling for distributed task (the underlying task may have succeeded)') from e
      if 'output' in task:
        return task['output']
      if 'exc' in task:
        raise OPTaskException(task['exc'])
      i = task['n_beats']


class TaskWorker:

  def __init__(self, manager, task_type_id, max_workers, heartbeat_s, fn):
    self._m = manager
    self._task_type_id = task_type_id
    self._heartbeat_s = heartbeat_s # Potentially None
    self._fn = fn
    self._thread_pool_executor = ThreadPoolExecutor(max_workers=max_workers)
    self._heartbeat_thread_pool_executor = ThreadPoolExecutor(max_workers=max_workers)
    self._n_in_progress = 0
    self._n_in_progress_lock = RLock()


  def _schedule_heartbeat(self, delay_s, task_run_id, task, lock, event):
    logger.debug('Scheduling heartbeat with %ss delay for task run ID: %s' % (delay_s, task_run_id))
    self._heartbeat_thread_pool_executor.submit(self._heartbeat, delay_s, task_run_id, task, lock, event)


  def _heartbeat(self, delay_s, task_run_id, task, heartbeat_lock, task_completed_event):
    if task_completed_event.wait(delay_s):
      return
    with heartbeat_lock:
      with self._m._exc_handler('Exception reporting heartbeat'):
        if task.get('end'):
          return
        logger.debug('Reporting heartbeat for task run ID: %s' % task_run_id)
        i = task['n_beats']
        next_delay_s = self._heartbeat_s or _exp(i, HEARTBEAT_EXP_GROWTH_FACTOR, MIN_HEARTBEAT_S, MAX_HEARTBEAT_S)
        task['deadline'] = _now() + timedelta(seconds=next_delay_s + HEARTBEAT_BUFFER_S)
        task['n_beats'] = i + 1
        self._m._service.db.update(self._m._data_type, task_run_id, task)
      self._schedule_heartbeat(next_delay_s, task_run_id, task, heartbeat_lock, task_completed_event)


  def _run(self, task_run_id, task, heartbeat_lock, task_completed_event):
    logger.debug("Beginning work on task run ID '%s' for task type ID '%s'" % (task_run_id, self._task_type_id))
    heartbeat_delay_s = self._heartbeat_s or MIN_HEARTBEAT_S
    task['n_beats'] = 0
    task['deadline'] = _now() + timedelta(seconds=heartbeat_delay_s + HEARTBEAT_BUFFER_S)
    self._schedule_heartbeat(heartbeat_delay_s, task_run_id, task, heartbeat_lock, task_completed_event)
    try:
      with self._m._parent(task_run_id):
        args, kwargs = pickle.loads(task['p_input'])
        if asyncio.iscoroutinefunction(self._fn):
          task['output'] = asyncio.run(self._fn(*args, **kwargs))
        else:
          task['output'] = self._fn(*args, **kwargs)
    except Exception as e:
      task['exc'] = e
      msg = "Exception for task type ID '%s'" % self._task_type_id
      logger.exception(msg)
      self._m._report_task_exception(msg, task_run_id, task)
    with heartbeat_lock:
      task['deadline'] = None
      task['end'] = _now()
      with self._m._exc_handler('Exception reporting task completion'):
        # If this call fails, the task will appear to be abandoned and get re-run at some point
        self._m._service.db.update(self._m._data_type, task_run_id, task)
    task_completed_event.set()
    with self._n_in_progress_lock:
      self._n_in_progress -= 1
    logger.debug("Finished work on task run ID '%s' for task type ID '%s'" % (task_run_id, self._task_type_id))


  def _execute_task(self, task_run_id, task):
    with self._n_in_progress_lock:
      self._n_in_progress += 1
      task_completed_event = Event()
      self._thread_pool_executor.submit(self._run, task_run_id, task, Lock(), task_completed_event)
    return LocalTaskResult(task, task_completed_event)


  @contextmanager
  def _get_capacity(self):
    with self._n_in_progress_lock:
      yield max(self._thread_pool_executor._max_workers - self._n_in_progress, 0)


  # TODO: expand this interface to allow for scheduled tasks where 't' is a timezone-aware datetime
  # def schedule(self, t, *args, **kwargs):
  #   fn.schedule(<datetime>, *args, **kwargs)


  # TODO: expose scheduled / compelted / errored tasks from the db via an API


  def submit(self, *args, **kwargs):
    parent_task_run_id = self._m._get_parent()
    logger.debug("Submitting task of type '%s' with parent '%s'" % (self._task_type_id, parent_task_run_id))
    pickled_input = pickle.dumps((args, kwargs))
    if parent_task_run_id:
      results = self._m._db.list(
        self._m._data_type,
        Field('parent') == parent_task_run_id,
        Field('task_type_id') == self._task_type_id,
        Field('p_input') == pickled_input
      )
      if results:
        task_run_id, task = results[0]
        if 'end' in task:
          logger.debug('Found completed child task for parent: %s' % parent_task_run_id)
          return TaskResult(task)
        else:
          logger.debug('Found task in progress for parent: %s' % parent_task_run_id)
          return DistributedTaskResult(task_run_id, self._m, self._heartbeat_s)

    new_task = {
      'source': OP_DEV_METADATA,
      'task_type_id': self._task_type_id,
      'parent': parent_task_run_id,
      'p_input': pickled_input,
      'start': _now(),
      'n_beats': 0,
      'deadline': _now() + timedelta(seconds=(self._heartbeat_s or MIN_HEARTBEAT_S) + HEARTBEAT_BUFFER_S)
    }
    with self._get_capacity() as capacity:
      if capacity > 0:
        task_run_id = self._m._service.db.create(self._m._data_type, new_task)
        logger.debug('Executing task for task type ID: %s' % self._task_type_id)
        return self._execute_task(task_run_id, new_task)
      else:
        # We didn't have room for this task, so we deliberately abandon it so it can get picked up by someone else ASAP
        new_task['deadline'] = _now()
        task_run_id = self._m._service.db.create(self._m._data_type, new_task)
        logger.debug('Deferring execution of new task for task type ID: %s' % self._task_type_id)
        return DistributedTaskResult(task_run_id, self._m, self._heartbeat_s)


  def __call__(self, *args, **kwargs):
    return self.submit(*args, **kwargs).result()


def _try_str(value):
  with suppress(Exception):
    return str(args)
  return '?'


# Builds a 'task' for input to the 'report_issue' alarm function
def _build_issue_task(task_run_id, task):
  args, kwargs = pickle.loads(task['p_input'])
  return {
    'task_run_id': task_run_id,
    'source': task['source'],
    'task_type_id': task['task_type_id'],
    'start': task['start'],
    'n_beats': task['n_beats'],
    'deadline': task['deadline'],
    'input': {
      'args': [ _try_str(x) for x in args ],
      'kwargs': { x: _try_str(y) for x, y in kwargs.items() }
    }
  }


class OPTaskManager:


  def __init__(self, service, abandoned_poll_interval_s=ABANDONED_POLL_INTERVAL_S, data_type='op_distributed_task'):
    self._service = service
    self._data_type = data_type

    # Note that we never un-register task workers
    self._registered_task_workers = {}
    self._registered_task_workers_lock = Lock()

    self._abandoned_poll_interval_s = abandoned_poll_interval_s
    self._thread_local = threading.local()

    # TODO: you still need to create these indexes...
    # service.db.init(
    #   data_type,
    #   computed=[ ComputedIndexField('is_completed', lambda x: 'end' in x) ],
    #   indexes=[ Index('is_completed', 'deadline'), Index('parent', 'task_type_id', 'p_input') ]
    # )

    # Note: there is some risk in starting this thread right away - we might find an abandoned task that has not YET
    # been registered, but may be in the future. However, we do have up to the time specified by the
    # 'abandoned_poll_interval_s' duration, which seems pretty safe.
    service.init(lambda: Thread(target=self._poll_abandoned_tasks, daemon=True).start())


  @contextmanager
  def _exc_handler(self, msg):
    with self._service.alarm.report_exceptions(msg, log=logger, metadata=OP_DEV_METADATA):
      yield


  # Should be called from an 'except' context
  def _report_task_exception(self, msg, task_run_id, task):
    _, exc_value, exc_traceback = sys.exc_info()
    stack_summary = traceback.extract_tb(exc_traceback)
    issue_exc = { 'msg': repr(exc_value), 'traceback': stack_summary.format() }
    issue_task = _build_issue_task(task_run_id, task)
    self._service.alarm.report_issue(msg, exception=issue_exc, task=issue_task, metadata=OP_DEV_METADATA)


  def _poll_abandoned_tasks(self):
    while True:
      # TODO: interrupt this sleep when a worker becomes available if we had no capacity available last we checked
      time.sleep(self._abandoned_poll_interval_s)
      logger.debug('Checking for abandoned tasks')
      with self._exc_handler('Exception handling abandoned tasks'):
        for task_run_id, task in self._db.iter(self._data_type, Field('is_completed') == False, Field('deadline') < _now()):
          # We found an abandoned task
          if task['task_type_id'] in self._registered_task_workers:
            # We know how to complete this task
            task_worker = self._registered_task_workers[task['task_type_id']]
            with task_worker._get_capacity() as capacity:
              if capacity > 0:
                delay_s = (task_worker._heartbeat_s or MIN_HEARTBEAT_S) + HEARTBEAT_BUFFER_S
                previous_deadline = task.get('deadline')
                task['deadline'] = _now() + timedelta(seconds=delay_s)
                task['n_beats'] = 0
                try:
                  self._db.update(self._data_type, task_run_id, task, Field('is_completed') == False, Field('deadline') == previous_deadline)
                except NotFoundException:
                  # This task has been picked up by someone else
                  logger.debug('Task already picked up with ID: %s' % task_run_id)
                  continue
                logger.debug('Executing abandoned task with ID: %s' % task_run_id)
                task_worker._execute_task(task_run_id, task)
              else:
                # We may not have capacity to take on any new tasks for this worker, but we might for some other
                # worker types
                # TODO: improve this log message - we should just log once at the end of our loop
                msg = 'No capacity available to execute abandoned tasks for task type ID: %s' % task['task_type_id']
                logger.debug(msg)
                continue
          else:
            # We don't know how to complete this task
            msg = _build_abandoned_task_msg(task)
            logger.critical(msg)
            issue_task = _build_issue_task(task_run_id, task)
            self._service.alarm.report_issue(msg, task=issue_task, metadata=OP_DEV_METADATA)


  @contextmanager
  def _parent(self, task_run_id):
    self._thread_local.parent = task_run_id
    try:
      yield
    finally:
      self._thread_local.parent = None


  def _get_parent(self):
    return getattr(self._thread_local, 'parent', None)


  def __call__(self, task_type_id=None, max_workers=64, heartbeat_s=None):
    def decorator(fn):
      nonlocal task_type_id, max_workers, heartbeat_s
      if not task_type_id:
        task_type_id = _generate_task_type_id(fn)
      with self._registered_task_workers_lock:
        if task_type_id in self._registered_task_workers:
          raise Exception('Task type already registered: %s' % task_type_id)
        task_worker = TaskWorker(self, task_type_id, max_workers, heartbeat_s, fn)
        self._registered_task_workers[task_type_id] = task_worker
      return task_worker
    return decorator



# TEST
# logging.basicConfig(level=logging.DEBUG)

# from op_storage import factory
# from uuid import UUID

# config = ...
# db = factory.build('mongo_db', config)
# # db._remove_data_type('op_distributed_task')
# task = OPTaskManager(db, { 'pkg_id': None, 'config_id': None, 'commit_id': None }, abandoned_poll_interval_s=2)


# def print_tasks():
#   print('--------------------')
#   for task_id, task in db.list('op_distributed_task'):
#     print(task['task_type_id'], task_id, task['start'].isoformat(), task['end'].isoformat() if 'end' in task else None, 'parent:', task['parent'])
#   print('--------------------')
# print_tasks()


# @task(task_type_id='workflow_1', heartbeat_s=1)
# def workflow_1(x_0):
#   print('workflow_1 start')
#   x_1 = step_1(x_0 + ':1')
#   x_2 = step_2(x_1 + ':2')
#   x_3 = step_3(x_2 + ':3')
#   print('workflow_1 finish')
#   return x_3

# @task(task_type_id='step_1', heartbeat_s=1)
# def step_1(x):
#   print('step_1 start')
#   time.sleep(5)
#   print('step_1 finish')
#   return x + ':step_1'

# @task(task_type_id='step_2', heartbeat_s=1)
# def step_2(x):
#   print('step_2 start')
#   time.sleep(5)
#   print('step_2 finish')
#   return x + ':step_2'

# @task(task_type_id='step_3', heartbeat_s=1)
# def step_3(x):
#   print('step_3 start')
#   time.sleep(5)
#   print('step_3 finish')
#   return x + ':step_3'

# # workflow_1.submit('A')
# time.sleep(20)



import random
import string


BAD_WORDS = ["ass", "cock", "crap", "cunt", "fuck", "nigger", "piss", "pussy", "shit", "tits"]
SERVICE_INSTANCE_ID_LEN = 8


def _generate_service_instance_id():
  pool = string.ascii_lowercase + string.digits
  service_instance_id = ""
  for i in range(SERVICE_INSTANCE_ID_LEN):
    service_instance_id += random.choice(pool)
  for bad_word in BAD_WORDS:
    if bad_word in service_instance_id:
      return _generate_service_instance_id()
  return service_instance_id


SERVICE_INSTANCE_ID = _generate_service_instance_id()
